"""Table and column models for Coda MCP server."""

from datetime import datetime
from typing import Literal, Union

from pydantic import Field

from .common import CodaBaseModel, FormulaDetail, PageReference

# Table Type Enum
TableType = Literal["table", "view"]


# Layout Types
Layout = Literal[
    "default",
    "areaChart",
    "barChart",
    "bubbleChart",
    "calendar",
    "card",
    "detail",
    "form",
    "ganttChart",
    "lineChart",
    "masterDetail",
    "pieChart",
    "scatterChart",
    "slide",
    "wordCloud",
]


# Sort Direction
SortDirection = Literal["ascending", "descending"]


# Column Format Types
ColumnFormatType = Literal[
    "text",
    "person",
    "lookup",
    "number",
    "percent",
    "currency",
    "date",
    "dateTime",
    "time",
    "duration",
    "email",
    "link",
    "slider",
    "scale",
    "image",
    "imageReference",
    "attachments",
    "button",
    "checkbox",
    "select",
    "packObject",
    "reaction",
    "canvas",
    "other",
]


# Currency Format Type
CurrencyFormatType = Literal["currency", "accounting", "financial"]


# Email Display Type
EmailDisplayType = Literal["iconAndEmail", "iconOnly", "emailOnly"]


# Link Display Type
LinkDisplayType = Literal["iconOnly", "url", "title", "card", "embed"]


# Image Shape Style
ImageShapeStyle = Literal["auto", "circle"]


# Slider Display Type
SliderDisplayType = Literal["slider", "progress"]


# Checkbox Display Type
CheckboxDisplayType = Literal["toggle", "check"]


# Duration Unit
DurationUnit = Literal["days", "hours", "minutes", "seconds"]


# Icon Set
IconSet = Literal[
    "star",
    "circle",
    "fire",
    "bug",
    "diamond",
    "bell",
    "thumbsup",
    "heart",
    "chili",
    "smiley",
    "lightning",
    "currency",
    "coffee",
    "person",
    "battery",
    "cocktail",
    "cloud",
    "sun",
    "checkmark",
    "lightbulb",
]


class NumberOrNumberFormula(CodaBaseModel):
    """A number or a string representing a formula that evaluates to a number."""

    value: float | str = Field(..., description="A numeric value or formula that evaluates to a numeric value.")


class ColumnReference(CodaBaseModel):
    """Reference to a column."""

    id: str = Field(..., description="ID of the column.", examples=["c-tuVwxYz"])
    type: Literal["column"] = Field(..., description="The type of this resource.")
    href: str = Field(
        ...,
        description="API link to the column.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/columns/c-tuVwxYz"],
    )


class TableReference(CodaBaseModel):
    """Reference to a table or view."""

    id: str = Field(..., description="ID of the table.", examples=["grid-pqRst-U"])
    type: Literal["table"] = Field(..., description="The type of this resource.")
    table_type: TableType = Field(..., description="Type of the table.")
    href: str = Field(
        ...,
        description="API link to the table.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U"],
    )
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the table.",
        examples=["https://coda.io/d/_dAbCDeFGH/#Teams-and-Tasks_tpqRst-U"],
    )
    name: str = Field(..., description="Name of the table.", examples=["Tasks"])
    parent: PageReference = Field(..., description="Parent page of the table.")


class Sort(CodaBaseModel):
    """A sort applied to a table or view."""

    column: ColumnReference = Field(..., description="Column to sort by.")
    direction: SortDirection = Field(..., description="Direction of the sort.")


class SimpleColumnFormat(CodaBaseModel):
    """Format of a simple column."""

    type: ColumnFormatType = Field(..., description="Format type of the column.")
    is_array: bool = Field(..., description="Whether or not this column is an array.", examples=[True])


class ReferenceColumnFormat(SimpleColumnFormat):
    """Format of a column that refers to another table."""

    table: TableReference = Field(..., description="Reference to the table this column refers to, if applicable.")


class NumericColumnFormat(SimpleColumnFormat):
    """Format of a numeric column."""

    precision: int | None = Field(None, ge=0, le=10, description="The decimal precision.", examples=[2])
    use_thousands_separator: bool | None = Field(
        None,
        description='Whether to use a thousands separator (like ",") to format the numeric value.',
        examples=[True],
    )


class CurrencyColumnFormat(SimpleColumnFormat):
    """Format of a currency column."""

    currency_code: str | None = Field(None, description="The currency symbol", examples=["$"])
    precision: int | None = Field(None, ge=0, le=10, description="The decimal precision.", examples=[2])
    format: CurrencyFormatType | None = Field(
        None,
        description="How the numeric value should be formatted (with or without symbol, negative numbers in parens).",
    )


class DateColumnFormat(SimpleColumnFormat):
    """Format of a date column."""

    format: str | None = Field(
        None,
        description="A format string using Moment syntax: https://momentjs.com/docs/#/displaying/",
        examples=["YYYY-MM-DD"],
    )


class TimeColumnFormat(SimpleColumnFormat):
    """Format of a time column."""

    format: str | None = Field(
        None,
        description="A format string using Moment syntax: https://momentjs.com/docs/#/displaying/",
        examples=["h:mm:ss A"],
    )


class DateTimeColumnFormat(SimpleColumnFormat):
    """Format of a date column."""

    date_format: str | None = Field(
        None,
        description="A format string using Moment syntax: https://momentjs.com/docs/#/displaying/",
        examples=["YYYY-MM-DD"],
    )
    time_format: str | None = Field(
        None,
        description="A format string using Moment syntax: https://momentjs.com/docs/#/displaying/",
        examples=["h:mm:ss A"],
    )


class DurationColumnFormat(SimpleColumnFormat):
    """Format of a duration column."""

    precision: int | None = Field(None, description="The precision.", examples=[2])
    max_unit: DurationUnit | None = Field(
        None,
        description='The maximum unit of precision, e.g. "hours" if this duration need not include minutes.',
    )


class EmailColumnFormat(SimpleColumnFormat):
    """Format of an email column."""

    display: EmailDisplayType | None = Field(
        None, description="How an email address should be displayed in the user interface."
    )
    autocomplete: bool | None = Field(None, description="Enable autocomplete for email.")


class LinkColumnFormat(SimpleColumnFormat):
    """Format of a link column."""

    display: LinkDisplayType | None = Field(None, description="How a link should be displayed in the user interface.")
    force: bool | None = Field(
        None,
        description="Force embeds to render on the client instead of the server (for sites that require user login).",
        examples=[True],
    )


class ImageReferenceColumnFormat(SimpleColumnFormat):
    """Format of an image reference column."""

    width: NumberOrNumberFormula = Field(..., description="The image width.")
    height: NumberOrNumberFormula = Field(..., description="The image height.")
    style: ImageShapeStyle = Field(..., description="How an image should be displayed.")


class SliderColumnFormat(SimpleColumnFormat):
    """Format of a numeric column that renders as a slider."""

    minimum: NumberOrNumberFormula | None = Field(None, description="The minimum allowed value for this slider.")
    maximum: NumberOrNumberFormula | None = Field(None, description="The maximum allowed value for this slider.")
    step: NumberOrNumberFormula | None = Field(None, description="The step size (numeric increment) for this slider.")
    display_type: SliderDisplayType | None = Field(None, description="How the slider should be rendered.")
    show_value: bool | None = Field(
        None,
        description="Whether the underyling numeric value is also displayed.",
        examples=[True],
    )


class ScaleColumnFormat(SimpleColumnFormat):
    """Format of a numeric column that renders as a scale, like star ratings."""

    maximum: float = Field(..., description="The maximum number allowed for this scale.", examples=[5])
    icon: IconSet = Field(..., description="The icon set to use when rendering the scale, e.g. render a 5 star scale.")


class ButtonColumnFormat(SimpleColumnFormat):
    """Format of a button column."""

    label: str | None = Field(None, description="Label formula for the button.", examples=["Click me"])
    disable_if: str | None = Field(None, description="DisableIf formula for the button.", examples=["False()"])
    action: str | None = Field(
        None,
        description="Action formula for the button.",
        examples=['OpenUrl("www.google.com")'],
    )


class CheckboxColumnFormat(SimpleColumnFormat):
    """Format of a checkbox column."""

    display_type: CheckboxDisplayType = Field(..., description="How a checkbox should be displayed.")


class SelectOption(CodaBaseModel):
    """An option for a select column."""

    name: str = Field(..., description="The name of the option.", examples=["Option 1"])
    background_color: str | None = Field(
        None,
        description="The background color of the option.",
        examples=["#ff0000"],
    )
    foreground_color: str | None = Field(
        None,
        description="The foreground color of the option.",
        examples=["#ffffff"],
    )


class SelectColumnFormat(SimpleColumnFormat):
    """Format of a select column."""

    pass


# Column Format Union Type
ColumnFormat = Union[
    ButtonColumnFormat,
    CheckboxColumnFormat,
    DateColumnFormat,
    DateTimeColumnFormat,
    DurationColumnFormat,
    EmailColumnFormat,
    LinkColumnFormat,
    CurrencyColumnFormat,
    ImageReferenceColumnFormat,
    NumericColumnFormat,
    ReferenceColumnFormat,
    SelectColumnFormat,
    SimpleColumnFormat,
    ScaleColumnFormat,
    SliderColumnFormat,
    TimeColumnFormat,
]


class Column(CodaBaseModel):
    """Info about a column."""

    id: str = Field(..., description="ID of the column.", examples=["c-tuVwxYz"])
    type: Literal["column"] = Field(..., description="The type of this resource.")
    href: str = Field(
        ...,
        description="API link to the column.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/columns/c-tuVwxYz"],
    )
    name: str = Field(..., description="Name of the column.", examples=["Completed"])
    display: bool | None = Field(None, description="Whether the column is the display column.", examples=[True])
    calculated: bool | None = Field(None, description="Whether the column has a formula set on it.", examples=[True])
    formula: str | None = Field(None, description="Formula on the column.", examples=["thisRow.Created()"])
    default_value: str | None = Field(None, description="Default value formula for the column.", examples=["Test"])
    format: ColumnFormat = Field(..., description="Format of the column.")


class ColumnDetail(Column):
    """Info about a column with parent table."""

    parent: TableReference = Field(..., description="Parent table of the column.")


class ColumnList(CodaBaseModel):
    """List of columns."""

    items: list[Column] = Field(..., description="Array of columns.")
    href: str | None = Field(
        None,
        description="API link to these results",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/columns?limit=20"],
    )
    next_page_token: str | None = Field(
        None,
        description="If specified, an opaque token used to fetch the next page of results.",
        examples=["eyJsaW1pd"],
    )
    next_page_link: str | None = Field(
        None,
        description="If specified, a link that can be used to fetch the next page of results.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/columns?pageToken=eyJsaW1pd"],
    )


class Table(CodaBaseModel):
    """Metadata about a table."""

    id: str = Field(..., description="ID of the table.", examples=["grid-pqRst-U"])
    type: Literal["table"] = Field(..., description="The type of this resource.")
    table_type: TableType = Field(..., description="Type of the table.")
    href: str = Field(
        ...,
        description="API link to the table.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U"],
    )
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the table.",
        examples=["https://coda.io/d/_dAbCDeFGH/#Teams-and-Tasks_tpqRst-U"],
    )
    name: str = Field(..., description="Name of the table.", examples=["Tasks"])
    parent: PageReference = Field(..., description="Parent page of the table.")
    parent_table: TableReference | None = Field(None, description="Parent table if this is a view.")
    display_column: ColumnReference = Field(..., description="The display column for the table.")
    row_count: int = Field(..., description="Total number of rows in the table.", examples=[130])
    sorts: list[Sort] = Field(..., description="Any sorts applied to the table.")
    layout: Layout = Field(..., description="Layout type of the table or view.")
    filter: FormulaDetail | None = Field(
        None, description="Detailed information about the filter formula for the table, if applicable."
    )
    created_at: datetime = Field(
        ...,
        description="Timestamp for when the table was created.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    updated_at: datetime = Field(
        ...,
        description="Timestamp for when the table was last modified.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    view_id: str | None = Field(None, description="ID of the view if this is a view.")


class TableList(CodaBaseModel):
    """List of tables."""

    items: list[TableReference] = Field(..., description="Array of table references.")
    href: str | None = Field(
        None,
        description="API link to these results",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables?limit=20"],
    )
    next_page_token: str | None = Field(
        None,
        description="If specified, an opaque token used to fetch the next page of results.",
        examples=["eyJsaW1pd"],
    )
    next_page_link: str | None = Field(
        None,
        description="If specified, a link that can be used to fetch the next page of results.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables?pageToken=eyJsaW1pd"],
    )



"""Row-related MCP tools for Coda tables."""

from typing import Literal

from ..client import CodaClient, clean_params
from ..models import (
    Method,
    Row,
    RowDeleteResult,
    RowEdit,
    RowList,
    RowsDelete,
    RowsDeleteResult,
    RowsUpsert,
    RowsUpsertResult,
    RowUpdate,
    RowUpdateResult,
)


async def list_rows(
    client: CodaClient,
    doc_id: str,
    table_id_or_name: str,
    query: str | None = None,
    sort_by: str | None = None,
    use_column_names: bool | None = None,
    value_format: Literal["simple", "simpleWithArrays", "rich"] | None = None,
    visible_only: bool | None = None,
    limit: int | None = None,
    page_token: str | None = None,
    sync_token: str | None = None,
) -> RowList:
    """List rows in a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        query: Query to filter rows (e.g., 'Status="Complete"').
        sort_by: Column to sort by. Use 'natural' for the table's sort order.
        use_column_names: Use column names instead of IDs in the response.
        value_format: Format for cell values (simple, simpleWithArrays, or rich).
        visible_only: If true, only return visible rows.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        sync_token: Token for incremental sync of changes.

    Returns:
        RowList with rows and pagination metadata.
    """
    params = {
        "query": query,
        "sortBy": sort_by,
        "useColumnNames": use_column_names,
        "valueFormat": value_format,
        "visibleOnly": visible_only,
        "limit": limit,
        "pageToken": page_token,
        "syncToken": sync_token,
    }
    result = await client.request(
        Method.GET, f"docs/{doc_id}/tables/{table_id_or_name}/rows", params=clean_params(params)
    )
    return RowList.model_validate(result)


async def get_row(
    client: CodaClient,
    doc_id: str,
    table_id_or_name: str,
    row_id_or_name: str,
    use_column_names: bool | None = None,
    value_format: Literal["simple", "simpleWithArrays", "rich"] | None = None,
) -> Row:
    """Get a specific row from a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row.
        use_column_names: Use column names instead of IDs in the response.
        value_format: Format for cell values (simple, simpleWithArrays, or rich).

    Returns:
        Row data with values.
    """
    params = {
        "useColumnNames": use_column_names,
        "valueFormat": value_format,
    }
    result = await client.request(
        Method.GET, f"docs/{doc_id}/tables/{table_id_or_name}/rows/{row_id_or_name}", params=clean_params(params)
    )
    return Row.model_validate(result)


async def upsert_rows(
    client: CodaClient,
    doc_id: str,
    table_id_or_name: str,
    rows: list[RowEdit],
    key_columns: list[str] | None = None,
    disable_parsing: bool | None = None,
) -> RowsUpsertResult:
    """Insert or update rows in a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        rows: List of rows to upsert. Each row should have a 'cells' array with column/value pairs.
        key_columns: Column IDs/names to use as keys for matching existing rows.
        disable_parsing: If true, cell values won't be parsed (e.g., URLs won't become links).

    Returns:
        RowsUpsertResult with the result of the upsert operation.
    """
    # Build the request model
    request = RowsUpsert(rows=rows, key_columns=key_columns)

    # Add disableParsing as a query parameter if provided
    params = {"disableParsing": disable_parsing} if disable_parsing is not None else None

    result = await client.request(
        Method.POST,
        f"docs/{doc_id}/tables/{table_id_or_name}/rows",
        json=request,
        params=clean_params(params) if params else None,
    )
    return RowsUpsertResult.model_validate(result)


async def update_row(
    client: CodaClient,
    doc_id: str,
    table_id_or_name: str,
    row_id_or_name: str,
    row: RowEdit,
    disable_parsing: bool | None = None,
) -> RowUpdateResult:
    """Update a specific row in a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row to update.
        row: Row data with cells array containing column/value pairs.
        disable_parsing: If true, cell values won't be parsed.

    Returns:
        RowUpdateResult with the updated row data.
    """
    # Build the request model
    request = RowUpdate(row=row)

    # Add disableParsing as a query parameter if provided
    params = {"disableParsing": disable_parsing} if disable_parsing is not None else None

    result = await client.request(
        Method.PUT,
        f"docs/{doc_id}/tables/{table_id_or_name}/rows/{row_id_or_name}",
        json=request,
        params=clean_params(params) if params else None,
    )
    return RowUpdateResult.model_validate(result)


async def delete_row(client: CodaClient, doc_id: str, table_id_or_name: str, row_id_or_name: str) -> RowDeleteResult:
    """Delete a specific row from a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row to delete.

    Returns:
        RowDeleteResult with the result of the deletion.
    """
    result = await client.request(Method.DELETE, f"docs/{doc_id}/tables/{table_id_or_name}/rows/{row_id_or_name}")
    return RowDeleteResult.model_validate(result)


async def delete_rows(
    client: CodaClient,
    doc_id: str,
    table_id_or_name: str,
    row_ids: list[str],
) -> RowsDeleteResult:
    """Delete multiple rows from a table.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_ids: List of row IDs to delete.

    Returns:
        RowsDeleteResult with the result of the deletion operation.
    """
    # Build the request model
    request = RowsDelete(row_ids=row_ids)

    result = await client.request(
        Method.DELETE,
        f"docs/{doc_id}/tables/{table_id_or_name}/rows",
        json=request,
    )
    return RowsDeleteResult.model_validate(result)



"""Tests for CodaClient HTTP request handling with mocked responses."""

import json

import pytest
from aioresponses import aioresponses

from coda_mcp_server.client import CodaClient
from coda_mcp_server.models import DocUpdate, Method


class TestClientRequestMethods:
    """Test CodaClient.request() method with various HTTP methods."""

    @pytest.mark.asyncio
    async def test_get_request_success(self, mock_client: CodaClient) -> None:
        """Test successful GET request with 200 response."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs/test-doc",
                payload={"id": "test-doc", "name": "Test Doc"},
            )

            result = await mock_client.request(Method.GET, "docs/test-doc")

            assert result["id"] == "test-doc"
            assert result["name"] == "Test Doc"

    @pytest.mark.asyncio
    async def test_post_request_success(self, mock_client: CodaClient) -> None:
        """Test successful POST request."""
        with aioresponses() as m:
            m.post(
                "https://coda.io/apis/v1/docs",
                payload={"id": "new-doc-123", "requestId": "req-456"},
            )

            result = await mock_client.request(
                Method.POST,
                "docs",
                json={"title": "New Doc"},
            )

            assert result["id"] == "new-doc-123"
            assert result["requestId"] == "req-456"

    @pytest.mark.asyncio
    async def test_put_request_success(self, mock_client: CodaClient) -> None:
        """Test successful PUT request."""
        with aioresponses() as m:
            m.put(
                "https://coda.io/apis/v1/docs/test-doc/pages/page-123",
                payload={"id": "page-123", "requestId": "req-789"},
            )

            result = await mock_client.request(
                Method.PUT,
                "docs/test-doc/pages/page-123",
                json={"name": "Updated Page"},
            )

            assert result["id"] == "page-123"

    @pytest.mark.asyncio
    async def test_patch_request_success(self, mock_client: CodaClient) -> None:
        """Test successful PATCH request."""
        with aioresponses() as m:
            m.patch(
                "https://coda.io/apis/v1/docs/test-doc",
                payload={"id": "test-doc", "requestId": "req-patch"},
            )

            result = await mock_client.request(
                Method.PATCH,
                "docs/test-doc",
                json={"title": "Patched Title"},
            )

            assert result["requestId"] == "req-patch"

    @pytest.mark.asyncio
    async def test_delete_request_with_204_no_content(self, mock_client: CodaClient) -> None:
        """Test DELETE request returns empty dict for 204 No Content."""
        with aioresponses() as m:
            m.delete(
                "https://coda.io/apis/v1/docs/test-doc",
                status=204,
            )

            result = await mock_client.request(Method.DELETE, "docs/test-doc")

            assert result == {}


class TestPydanticModelSerialization:
    """Test auto-serialization of Pydantic models to camelCase for API."""

    @pytest.mark.asyncio
    async def test_pydantic_model_auto_serialization(self, mock_client: CodaClient) -> None:
        """Test that Pydantic models are automatically serialized to camelCase."""
        with aioresponses() as m:
            m.patch(
                "https://coda.io/apis/v1/docs/test-doc",
                payload={"id": "test-doc", "requestId": "req-123"},
            )

            # Create model with snake_case fields
            doc_update = DocUpdate(
                title="New Title",
                icon_name="rocket",  # snake_case
            )

            await mock_client.request(
                Method.PATCH,
                "docs/test-doc",
                json=doc_update,  # Pass Pydantic model
            )

            # Verify request was made
            assert len(m.requests) > 0

            # Get the request details - aioresponses stores body in different keys
            requests_list = list(m.requests.values())
            assert len(requests_list) > 0
            call = requests_list[0][0]

            # Try to get the body from various possible locations
            if "json" in call.kwargs:
                body = call.kwargs["json"]
            elif "data" in call.kwargs:
                body = json.loads(call.kwargs["data"])
            else:
                # Body might be in the request itself
                if hasattr(call, "body") and call.body:
                    body = json.loads(call.body)
                else:
                    # Skip detailed validation if we can't access the body
                    # The important thing is the model was accepted
                    pytest.skip("Cannot access request body in this test environment")

            # Should have camelCase in request
            assert "iconName" in body
            assert "icon_name" not in body
            assert body["iconName"] == "rocket"


class TestErrorHandling:
    """Test error handling for various HTTP error scenarios."""

    @pytest.mark.asyncio
    async def test_rate_limit_429(self, mock_client: CodaClient) -> None:
        """Test 429 rate limit error handling."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs",
                status=429,
                headers={"Retry-After": "60"},
            )

            with pytest.raises(Exception) as exc_info:
                await mock_client.request(Method.GET, "docs")

            assert "Rate limit exceeded" in str(exc_info.value)
            assert "60 seconds" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_404_not_found(self, mock_client: CodaClient) -> None:
        """Test 404 not found error."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs/nonexistent",
                status=404,
                payload={"message": "Doc not found"},
            )

            with pytest.raises(Exception) as exc_info:
                await mock_client.request(Method.GET, "docs/nonexistent")

            assert "API Error 404" in str(exc_info.value)
            assert "Doc not found" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_500_server_error(self, mock_client: CodaClient) -> None:
        """Test 500 internal server error."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs",
                status=500,
                payload={"error": "Internal server error"},
            )

            with pytest.raises(Exception) as exc_info:
                await mock_client.request(Method.GET, "docs")

            assert "API Error 500" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_invalid_json_response(self, mock_client: CodaClient) -> None:
        """Test handling of invalid JSON response."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs",
                status=200,
                body="Not valid JSON{",
            )

            with pytest.raises(Exception) as exc_info:
                await mock_client.request(Method.GET, "docs")

            assert "Invalid JSON" in str(exc_info.value)


class TestResponseParsing:
    """Test response parsing edge cases."""

    @pytest.mark.asyncio
    async def test_empty_response_body(self, mock_client: CodaClient) -> None:
        """Test handling of empty response body."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs",
                status=200,
                body="",
            )

            result = await mock_client.request(Method.GET, "docs")

            assert result == {}

    @pytest.mark.asyncio
    async def test_query_parameters(self, mock_client: CodaClient) -> None:
        """Test that query parameters are properly passed."""
        with aioresponses() as m:
            m.get(
                "https://coda.io/apis/v1/docs?limit=10&query=test",
                payload={"items": []},
            )

            result = await mock_client.request(
                Method.GET,
                "docs",
                params={"limit": 10, "query": "test"},
            )

            assert "items" in result



"""Row and cell value models for Coda MCP server."""

from datetime import datetime
from typing import Literal, Union

from pydantic import Field

from .common import CodaBaseModel, DocumentMutateResponse, PersonValue
from .tables import TableReference

# Value Format Types
ValueFormat = Literal["simple", "simpleWithArrays", "rich"]


# Rows Sort By
RowsSortBy = Literal["createdAt", "natural", "updatedAt"]


# Image Status
ImageStatus = Literal["live", "deleted", "failed"]


# Linked Data Types
LinkedDataType = Literal["ImageObject", "MonetaryAmount", "Person", "WebPage", "StructuredValue"]


# Scalar Value - primitive types
ScalarValue = Union[str, float, bool]


# Value - scalar or array of scalars
Value = Union[ScalarValue, list[Union[ScalarValue, list[ScalarValue]]]]


# Currency Amount
CurrencyAmount = Union[str, float]


class LinkedDataObject(CodaBaseModel):
    """Base type for a JSON-LD (Linked Data) object."""

    context: str = Field(
        ...,
        alias="@context",
        description='A url describing the schema context for this object, typically "http://schema.org/".',
        examples=["http://schema.org/"],
    )
    type: LinkedDataType = Field(..., alias="@type", description="A schema.org identifier for the object.")
    additional_type: str | None = Field(
        None,
        description=(
            "An identifier of additional type info specific to Coda that may not be present in a schema.org taxonomy."
        ),
    )


class UrlValue(LinkedDataObject):
    """A named hyperlink to an arbitrary url."""

    type: Literal["WebPage"] = Field(..., alias="@type", description="The type of this resource.")
    name: str | None = Field(None, description="The user-visible text of the hyperlink.", examples=["Click me"])
    url: str = Field(..., description="The url of the hyperlink.", examples=["https://coda.io"])


class ImageUrlValue(LinkedDataObject):
    """A named url of an image along with metadata."""

    type: Literal["ImageObject"] = Field(..., alias="@type", description="The type of this resource.")
    name: str | None = Field(None, description="The name of the image.", examples=["Dogs Playing Poker"])
    url: str | None = Field(
        None,
        description="The url of the image.",
        examples=["https://example.com/dogs-playing-poker.jpg"],
    )
    height: float | None = Field(None, description="The height of the image in pixels.", examples=[480])
    width: float | None = Field(None, description="The width of the image in pixels.", examples=[640])
    status: ImageStatus | None = Field(None, description="The status of the image.")


class CurrencyValue(LinkedDataObject):
    """A monetary value with its associated currency code."""

    type: Literal["MonetaryAmount"] = Field(..., alias="@type", description="The type of this resource.")
    currency: str = Field(..., description="The 3-letter currency code.", examples=["USD"])
    amount: CurrencyAmount = Field(..., description="The monetary amount as a string or number.", examples=["12.99"])


class RowValue(LinkedDataObject):
    """A value representing a Coda row."""

    type: Literal["StructuredValue"] = Field(..., alias="@type", description="The type of this resource.")
    name: str = Field(
        ...,
        description="The display name of the row, based on its identifying column.",
        examples=["Apple"],
    )
    url: str = Field(
        ...,
        description="The url of the row.",
        examples=["https://coda.io/d/_dAbCDeFGH#Teams-and-Tasks_tpqRst-U/_rui-tuVwxYz"],
    )
    table_id: str = Field(..., description="The ID of the table", examples=["grid-pqRst-U"])
    row_id: str = Field(..., description="The ID of the row", examples=["i-tuVwxYz"])
    table_url: str = Field(
        ...,
        description="The url of the table.",
        examples=["https://coda.io/d/_dAbCDeFGH#Teams-and-Tasks_tpqRst-U"],
    )
    additional_type: Literal["row"] = Field(..., description="The type of this resource.")


# Rich Single Value - scalar or structured data
RichSingleValue = Union[ScalarValue, CurrencyValue, ImageUrlValue, PersonValue, UrlValue, RowValue]


# Rich Value - single value or array of values
RichValue = Union[RichSingleValue, list[Union[RichSingleValue, list[RichSingleValue]]]]


# Cell Value - can be simple value or rich value
CellValue = Union[Value, RichValue]


class CellEdit(CodaBaseModel):
    """An edit made to a particular cell in a row."""

    column: str = Field(
        ...,
        description="Column ID, URL, or name (fragile and discouraged) associated with this edit.",
        examples=["c-tuVwxYz"],
    )
    value: Value = Field(..., description="The value to set in the cell.")


class RowEdit(CodaBaseModel):
    """An edit made to a particular row."""

    cells: list[CellEdit] = Field(..., description="Array of cell edits for the row.")


class Row(CodaBaseModel):
    """Info about a row."""

    id: str = Field(..., description="ID of the row.", examples=["i-tuVwxYz"])
    type: Literal["row"] = Field(..., description="The type of this resource.")
    href: str = Field(
        ...,
        description="API link to the row.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/rows/i-RstUv-W"],
    )
    name: str = Field(
        ...,
        description="The display name of the row, based on its identifying column.",
        examples=["Apple"],
    )
    index: int = Field(..., description="Index of the row within the table.", examples=[7])
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the row.",
        examples=["https://coda.io/d/_dAbCDeFGH#Teams-and-Tasks_tpqRst-U/_rui-tuVwxYz"],
    )
    created_at: datetime = Field(
        ...,
        description="Timestamp for when the row was created.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    updated_at: datetime = Field(
        ...,
        description="Timestamp for when the row was last modified.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    values: dict[str, CellValue] = Field(
        ...,
        description=(
            "Values for a specific row, represented as a hash of column IDs (or names with `useColumnNames`) to values."
        ),
        examples=[{"c-tuVwxYz": "Apple", "c-bCdeFgh": ["$12.34", "$56.78"]}],
    )


class RowDetail(Row):
    """Details about a row."""

    parent: TableReference = Field(..., description="Parent table of the row.")


class RowList(CodaBaseModel):
    """List of rows."""

    items: list[Row] = Field(..., description="Array of rows.")
    href: str | None = Field(
        None,
        description="API link to these results",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/rows?limit=20"],
    )
    next_page_token: str | None = Field(
        None,
        description="If specified, an opaque token used to fetch the next page of results.",
        examples=["eyJsaW1pd"],
    )
    next_page_link: str | None = Field(
        None,
        description="If specified, a link that can be used to fetch the next page of results.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/tables/grid-pqRst-U/rows?pageToken=eyJsaW1pd"],
    )
    next_sync_token: str | None = Field(
        None,
        description=(
            "If specified, an opaque token that can be passed back later to retrieve new results that match "
            "the parameters specified when the sync token was created."
        ),
        examples=["eyJsaW1pd"],
    )


class RowUpdate(CodaBaseModel):
    """Payload for updating a row in a table."""

    row: RowEdit = Field(..., description="The row data to update.")


class RowUpdateResult(DocumentMutateResponse):
    """The result of a row update."""

    id: str = Field(..., description="ID of the updated row.", examples=["i-tuVwxYz"])


class RowsUpsert(CodaBaseModel):
    """Payload for upserting rows in a table."""

    rows: list[RowEdit] = Field(..., description="Array of row edits to upsert.")
    key_columns: list[str] | None = Field(
        None,
        description=(
            "Optional column IDs, URLs, or names (fragile and discouraged), "
            "specifying columns to be used as upsert keys."
        ),
        examples=[["c-bCdeFgh"]],
    )


class RowsUpsertResult(DocumentMutateResponse):
    """The result of a rows insert/upsert operation."""

    added_row_ids: list[str] | None = Field(
        None,
        description=("Row IDs for rows that will be added. Only applicable when keyColumns is not set or empty."),
        examples=[["i-bCdeFgh", "i-CdEfgHi"]],
    )


class RowsDelete(CodaBaseModel):
    """Payload for deleting rows from a table."""

    row_ids: list[str] = Field(
        ...,
        description="Row IDs to delete.",
        examples=[["i-bCdeFgh", "i-CdEfgHi"]],
    )


class RowsDeleteResult(DocumentMutateResponse):
    """The result of a rows delete operation."""

    row_ids: list[str] = Field(
        ...,
        description="Row IDs to delete.",
        examples=[["i-bCdeFgh", "i-CdEfgHi"]],
    )


class RowDeleteResult(DocumentMutateResponse):
    """The result of a row deletion."""

    id: str = Field(..., description="ID of the row to be deleted.", examples=["i-tuVwxYz"])


class PushButtonResult(DocumentMutateResponse):
    """The result of a push button."""

    row_id: str = Field(..., description="ID of the row where the button exists.", examples=["i-tuVwxYz"])
    column_id: str = Field(..., description="ID of the column where the button exists.", examples=["i-tuVwxYz"])


# For convenience, export commonly used types for row creation
# RowCreate is effectively the same as RowEdit in the API
RowCreate = RowEdit



"""Pydantic models for Coda API."""

# Import common types and utilities
from .common import (
    AccessType,
    # Error models
    ApiError,
    BadRequestError,
    CodaDetail,
    DocReference,
    # Response models
    DocumentMutateResponse,
    FolderReference,
    ForbiddenError,
    FormulaDetail,
    GoneError,
    Icon,
    Image,
    # Enums
    Method,
    NotFoundError,
    # Reference models
    PageReference,
    # Pagination
    PaginationMetadata,
    PersonValue,
    SortBy,
    SortDirection,
    TableType,
    TooManyRequestsError,
    UnauthorizedError,
    UnprocessableEntityError,
    User,
    ValidationError,
    ValueFormat,
    WorkspaceReference,
)

# Import doc-related models
from .docs import (
    Doc,
    DocCategory,
    DocCategoryList,
    DocCreate,
    DocDelete,
    DocList,
    DocPublish,
    DocPublished,
    DocPublishMode,
    DocSize,
    DocumentCreationResult,
    DocUpdate,
    DocUpdateResult,
    PublishResult,
    UnpublishResult,
)

# Import page export models
from .exports import (
    BeginPageContentExportRequest,
    BeginPageContentExportResponse,
    PageContentExportStatusResponse,
)

# Import formula models
from .formulas import (
    Formula,
    FormulaList,
    FormulaReference,
)

# Import page-related models
from .pages import (
    CanvasPageContent,
    EmbedPageContent,
    Page,
    PageContent,
    PageContentUpdate,
    PageCreate,
    PageCreateResult,
    PageDeleteResult,
    PageList,
    PageUpdate,
    PageUpdateResult,
)

# Import row-related models
from .rows import (
    CellEdit,
    CellValue,
    CurrencyValue,
    ImageUrlValue,
    LinkedDataObject,
    PushButtonResult,
    RichSingleValue,
    RichValue,
    Row,
    RowCreate,
    RowDeleteResult,
    RowDetail,
    RowEdit,
    RowList,
    RowsDelete,
    RowsDeleteResult,
    RowsSortBy,
    RowsUpsert,
    RowsUpsertResult,
    RowUpdate,
    RowUpdateResult,
    RowValue,
    ScalarValue,
    UrlValue,
    Value,
)

# Import table-related models
from .tables import (
    ButtonColumnFormat,
    CheckboxColumnFormat,
    Column,
    ColumnDetail,
    ColumnFormat,
    ColumnFormatType,
    ColumnList,
    ColumnReference,
    CurrencyColumnFormat,
    DateColumnFormat,
    DateTimeColumnFormat,
    DurationColumnFormat,
    EmailColumnFormat,
    ImageReferenceColumnFormat,
    Layout,
    LinkColumnFormat,
    NumericColumnFormat,
    ReferenceColumnFormat,
    ScaleColumnFormat,
    SelectColumnFormat,
    SelectOption,
    SimpleColumnFormat,
    SliderColumnFormat,
    Sort,
    Table,
    TableList,
    TableReference,
    TimeColumnFormat,
)

__all__ = [
    # Common - Enums
    "Method",
    "AccessType",
    "SortDirection",
    "SortBy",
    "ValueFormat",
    "TableType",
    # Common - Error models
    "ApiError",
    "ValidationError",
    "CodaDetail",
    "BadRequestError",
    "UnauthorizedError",
    "ForbiddenError",
    "NotFoundError",
    "GoneError",
    "UnprocessableEntityError",
    "TooManyRequestsError",
    # Common - Pagination
    "PaginationMetadata",
    # Common - Reference models
    "PageReference",
    "Icon",
    "Image",
    "FormulaDetail",
    "PersonValue",
    "DocReference",
    "FolderReference",
    "WorkspaceReference",
    "User",
    # Common - Response models
    "DocumentMutateResponse",
    # Docs
    "Doc",
    "DocList",
    "DocCreate",
    "DocUpdate",
    "DocUpdateResult",
    "DocumentCreationResult",
    "DocDelete",
    "DocPublish",
    "DocPublished",
    "DocSize",
    "DocCategory",
    "DocCategoryList",
    "PublishResult",
    "UnpublishResult",
    "DocPublishMode",
    # Tables
    "Layout",
    "ColumnFormatType",
    "TableReference",
    "ColumnReference",
    "Column",
    "ColumnDetail",
    "ColumnList",
    "Table",
    "TableList",
    "Sort",
    "SimpleColumnFormat",
    "ReferenceColumnFormat",
    "NumericColumnFormat",
    "CurrencyColumnFormat",
    "DateColumnFormat",
    "TimeColumnFormat",
    "DateTimeColumnFormat",
    "DurationColumnFormat",
    "EmailColumnFormat",
    "LinkColumnFormat",
    "ImageReferenceColumnFormat",
    "SliderColumnFormat",
    "ScaleColumnFormat",
    "ButtonColumnFormat",
    "CheckboxColumnFormat",
    "SelectColumnFormat",
    "SelectOption",
    "ColumnFormat",
    # Rows
    "RowsSortBy",
    "ScalarValue",
    "Value",
    "CellValue",
    "CellEdit",
    "RowEdit",
    "Row",
    "RowDetail",
    "RowList",
    "RowUpdate",
    "RowUpdateResult",
    "RowCreate",
    "RowsUpsert",
    "RowsUpsertResult",
    "RowsDelete",
    "RowsDeleteResult",
    "RowDeleteResult",
    "PushButtonResult",
    "LinkedDataObject",
    "UrlValue",
    "ImageUrlValue",
    "CurrencyValue",
    "RowValue",
    "RichSingleValue",
    "RichValue",
    # Pages
    "Page",
    "PageList",
    "PageContent",
    "CanvasPageContent",
    "EmbedPageContent",
    "PageCreate",
    "PageContentUpdate",
    "PageUpdate",
    "PageCreateResult",
    "PageUpdateResult",
    "PageDeleteResult",
    # Exports
    "BeginPageContentExportRequest",
    "BeginPageContentExportResponse",
    "PageContentExportStatusResponse",
    # Formulas
    "Formula",
    "FormulaList",
    "FormulaReference",
]



"""Pydantic models for Coda doc-related API resources.

This module contains models for working with Coda docs, including:
- Doc: Full metadata about a doc
- DocList: List of docs with pagination
- DocCreate: Payload for creating a new doc
- DocUpdate: Payload for updating a doc
- DocDelete: Result of doc deletion
- DocPublish: Payload for publishing a doc
- DocPublished: Published doc information
- DocSize: Information about doc size/components
- PublishResult, UnpublishResult: Results of publish operations
- DocumentCreationResult: Result of doc creation
- DocUpdateResult: Result of doc update
"""

from typing import Literal

from pydantic import Field

from .common import (
    CodaBaseModel,
    DocReference,
    DocumentMutateResponse,
    FolderReference,
    Icon,
    WorkspaceReference,
)
from .pages import PageCreate


class DocSize(CodaBaseModel):
    """The number of components within a Coda doc.

    Provides information about the size and complexity of a doc.
    """

    total_row_count: float = Field(
        ...,
        description="The number of rows contained within all tables of the doc.",
        examples=[31337],
    )
    table_and_view_count: float = Field(
        ...,
        description="The total number of tables and views contained within the doc.",
        examples=[42],
    )
    page_count: float = Field(
        ...,
        description="The total number of page contained within the doc.",
        examples=[10],
    )
    over_api_size_limit: bool = Field(
        ...,
        description="If true, indicates that the doc is over the API size limit.",
        examples=[False],
    )


class DocCategory(CodaBaseModel):
    """The category applied to a doc.

    Categories are used to classify published docs in the Coda gallery.
    """

    name: str = Field(
        ...,
        description="Name of the category.",
        examples=["Project Management"],
    )


class DocPublishMode(str):
    """Which interaction mode the published doc should use.

    - view: Users can only view the doc
    - play: Users can interact with controls but not edit
    - edit: Users can fully edit the doc
    """

    VIEW = "view"
    PLAY = "play"
    EDIT = "edit"


class DocPublished(CodaBaseModel):
    """Information about the publishing state of the document.

    Contains details about how a doc is published and shared publicly.
    """

    description: str | None = Field(
        None,
        description="Description of the published doc.",
        examples=["Hello World!"],
    )
    browser_link: str = Field(
        ...,
        description="URL to the published doc.",
        examples=["https://coda.io/@coda/hello-world"],
    )
    image_link: str | None = Field(
        None,
        description="URL to the cover image for the published doc.",
    )
    discoverable: bool = Field(
        ...,
        description="If true, indicates that the doc is discoverable.",
        examples=[True],
    )
    earn_credit: bool = Field(
        ...,
        description=(
            "If true, new users may be required to sign in to view content within this document. "
            "You will receive Coda credit for each user who signs up via your doc."
        ),
        examples=[True],
    )
    mode: Literal["view", "play", "edit"] = Field(
        ...,
        description="Which interaction mode the published doc should use.",
    )
    categories: list[DocCategory] = Field(
        ...,
        description="Categories applied to the doc.",
        examples=[["Project Management"]],
    )


class Doc(CodaBaseModel):
    """Metadata about a Coda doc.

    Contains comprehensive information about a doc including ownership,
    timestamps, location in workspace/folder hierarchy, and publishing state.
    """

    id: str = Field(
        ...,
        description="ID of the Coda doc.",
        examples=["AbCDeFGH"],
    )
    type: Literal["doc"] = Field(
        ...,
        description="The type of this resource.",
    )
    href: str = Field(
        ...,
        description="API link to the Coda doc.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH"],
    )
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the Coda doc.",
        examples=["https://coda.io/d/_dAbCDeFGH"],
    )
    icon: Icon | None = Field(
        None,
        description="Info about the icon.",
    )
    name: str = Field(
        ...,
        description="Name of the doc.",
        examples=["Product Launch Hub"],
    )
    owner: str = Field(
        ...,
        description="Email address of the doc owner.",
        examples=["user@example.com"],
    )
    owner_name: str = Field(
        ...,
        description="Name of the doc owner.",
        examples=["Some User"],
    )
    doc_size: DocSize | None = Field(
        None,
        description="The number of components within a Coda doc.",
    )
    source_doc: DocReference | None = Field(
        None,
        description="Reference to a Coda doc from which this doc was copied, if any.",
    )
    created_at: str = Field(
        ...,
        description="Timestamp for when the doc was created.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    updated_at: str = Field(
        ...,
        description="Timestamp for when the doc was last modified.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    published: DocPublished | None = Field(
        None,
        description="Information about the publishing state of the document.",
    )
    folder: FolderReference = Field(
        ...,
        description="Reference to the folder containing this doc.",
    )
    workspace: WorkspaceReference = Field(
        ...,
        description="Reference to the workspace containing this doc.",
    )
    workspace_id: str = Field(
        ...,
        description="ID of the Coda workspace containing this doc.",
        examples=["ws-1Ab234"],
        deprecated=True,
    )
    folder_id: str = Field(
        ...,
        description="ID of the Coda folder containing this doc.",
        examples=["fl-1Ab234"],
        deprecated=True,
    )


class DocList(CodaBaseModel):
    """List of Coda docs.

    Contains a paginated list of docs with optional pagination tokens
    to fetch additional results.
    """

    items: list[Doc] = Field(
        ...,
        description="Array of Coda docs.",
    )
    href: str | None = Field(
        None,
        description="API link to these results",
        examples=["https://coda.io/apis/v1/docs?limit=20"],
    )
    next_page_token: str | None = Field(
        None,
        description="If specified, an opaque token used to fetch the next page of results.",
        examples=["eyJsaW1pd"],
    )
    next_page_link: str | None = Field(
        None,
        description="If specified, a link that can be used to fetch the next page of results.",
        examples=["https://coda.io/apis/v1/docs?pageToken=eyJsaW1pd"],
    )


class DocCreate(CodaBaseModel):
    """Payload for creating a new doc.

    A new doc can be created from scratch or by copying an existing doc.
    You can optionally specify a folder/workspace location and initial page content.
    """

    title: str | None = Field(
        None,
        description="Title of the new doc. Defaults to 'Untitled'.",
        examples=["Project Tracker"],
    )
    source_doc: str | None = Field(
        None,
        description="An optional doc ID from which to create a copy.",
        examples=["iJKlm_noPq"],
    )
    timezone: str | None = Field(
        None,
        description="The timezone to use for the newly created doc.",
        examples=["America/Los_Angeles"],
    )
    folder_id: str | None = Field(
        None,
        description=(
            'The ID of the folder within which to create this doc. Defaults to your "My docs" folder in the '
            "oldest workspace you joined; this is subject to change. You can get this ID by opening the folder "
            "in the docs list on your computer and grabbing the `folderId` query parameter."
        ),
        examples=["fl-ABcdEFgHJi"],
    )
    initial_page: PageCreate | None = Field(
        None,
        description="The contents of the initial page of the doc.",
    )


class DocUpdate(CodaBaseModel):
    """Payload for updating a doc.

    Allows updating basic doc properties like title and icon.
    """

    title: str | None = Field(
        None,
        description="Title of the doc.",
        examples=["Project Tracker"],
    )
    icon_name: str | None = Field(
        None,
        description="Name of the icon.",
        examples=["rocket"],
    )


class DocDelete(CodaBaseModel):
    """The result of a doc deletion.

    Returned when a doc is successfully deleted.
    """

    pass


class DocPublish(CodaBaseModel):
    """Payload for publishing a doc or updating its publishing information.

    Controls how the doc appears when published and who can access it.
    """

    slug: str | None = Field(
        None,
        description="Slug for the published doc.",
        examples=["my-doc"],
    )
    discoverable: bool | None = Field(
        None,
        description="If true, indicates that the doc is discoverable.",
        examples=[True],
    )
    earn_credit: bool | None = Field(
        None,
        description=(
            "If true, new users may be required to sign in to view content within this document. "
            "You will receive Coda credit for each user who signs up via your doc."
        ),
        examples=[True],
    )
    category_names: list[str] | None = Field(
        None,
        description="The names of categories to apply to the document.",
        examples=[["Project management"]],
    )
    mode: Literal["view", "play", "edit"] | None = Field(
        None,
        description="Which interaction mode the published doc should use.",
    )


class PublishResult(DocumentMutateResponse):
    """The result of publishing a doc.

    Extends DocumentMutateResponse with any publish-specific fields.
    """

    pass


class UnpublishResult(CodaBaseModel):
    """The result of unpublishing a doc.

    Returned when a doc is successfully unpublished.
    """

    pass


class DocumentCreationResult(CodaBaseModel):
    """The result of a doc creation.

    Similar to Doc but includes a requestId for tracking the creation operation.
    """

    id: str = Field(
        ...,
        description="ID of the Coda doc.",
        examples=["AbCDeFGH"],
    )
    type: Literal["doc"] = Field(
        ...,
        description="The type of this resource.",
    )
    href: str = Field(
        ...,
        description="API link to the Coda doc.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH"],
    )
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the Coda doc.",
        examples=["https://coda.io/d/_dAbCDeFGH"],
    )
    icon: Icon | None = Field(
        None,
        description="Info about the icon.",
    )
    name: str = Field(
        ...,
        description="Name of the doc.",
        examples=["Product Launch Hub"],
    )
    owner: str = Field(
        ...,
        description="Email address of the doc owner.",
        examples=["user@example.com"],
    )
    owner_name: str = Field(
        ...,
        description="Name of the doc owner.",
        examples=["Some User"],
    )
    doc_size: DocSize | None = Field(
        None,
        description="The number of components within a Coda doc.",
    )
    source_doc: DocReference | None = Field(
        None,
        description="Reference to a Coda doc from which this doc was copied, if any.",
    )
    created_at: str = Field(
        ...,
        description="Timestamp for when the doc was created.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    updated_at: str = Field(
        ...,
        description="Timestamp for when the doc was last modified.",
        examples=["2018-04-11T00:18:57.946Z"],
    )
    published: DocPublished | None = Field(
        None,
        description="Information about the publishing state of the document.",
    )
    folder: FolderReference = Field(
        ...,
        description="Reference to the folder containing this doc.",
    )
    workspace: WorkspaceReference = Field(
        ...,
        description="Reference to the workspace containing this doc.",
    )
    workspace_id: str = Field(
        ...,
        description="ID of the Coda workspace containing this doc.",
        examples=["ws-1Ab234"],
        deprecated=True,
    )
    folder_id: str = Field(
        ...,
        description="ID of the Coda folder containing this doc.",
        examples=["fl-1Ab234"],
        deprecated=True,
    )
    request_id: str = Field(
        ...,
        description="An arbitrary unique identifier for this request.",
        examples=["abc-123-def-456"],
    )


class DocUpdateResult(CodaBaseModel):
    """The result of a doc update.

    Returned when a doc is successfully updated.
    """

    pass


class DocCategoryList(CodaBaseModel):
    """A list of categories that can be applied to a doc.

    Used to discover available categories for publishing docs.
    """

    items: list[DocCategory] = Field(
        ...,
        description="Categories for the doc.",
    )



"""Pydantic models for Coda pages."""

from typing import Literal

from pydantic import Field

from .common import CodaBaseModel, DocumentMutateResponse, Icon, Image, PageReference


class Page(CodaBaseModel):
    """Metadata about a page."""

    id: str = Field(..., description="ID of the page.")
    type: Literal["page"] = Field(..., description="The type of this resource.")
    href: str = Field(..., description="API link to the page.")
    browser_link: str = Field(
        ...,
        description="Browser-friendly link to the page.",
        examples=["https://coda.io/d/_dAbCDeFGH/Launch-Status_sumnO"],
    )
    name: str = Field(..., description="Name of the page.", examples=["Launch Status"])
    subtitle: str | None = Field(
        None, description="Subtitle of the page.", examples=["See the status of launch-related tasks."]
    )
    icon: Icon | None = Field(None, description="Icon for the page.")
    image: Image | None = Field(None, description="Cover image for the page.")
    content_type: Literal["canvas", "embed", "syncPage"] = Field(..., description="The type of content on the page.")
    is_hidden: bool = Field(..., description="Whether the page is hidden in the UI.", examples=[True])
    is_effectively_hidden: bool = Field(
        ...,
        description="Whether the page or any of its parents is hidden in the UI.",
        examples=[True],
    )
    parent: PageReference | None = Field(None, description="Reference to the parent page.")
    children: list[PageReference] = Field(..., description="Child pages of this page.")


class PageList(CodaBaseModel):
    """List of pages."""

    items: list[Page] = Field(..., description="List of pages.")
    href: str | None = Field(
        None,
        description="API link to these results.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/pages?limit=20"],
    )
    next_page_token: str | None = Field(
        None, description="Token for fetching the next page of results.", examples=["eyJsaW1pd"]
    )
    next_page_link: str | None = Field(
        None,
        description="Link to the next page of results.",
        examples=["https://coda.io/apis/v1/docs/AbCDeFGH/pages?pageToken=eyJsaW1pd"],
    )


# ============================================================================
# Page Content Models (matches OpenAPI spec exactly)
# ============================================================================


class PageContent(CodaBaseModel):
    """Content for a page (canvas).

    Raw content with format and content string. This is the base content type
    referenced by other content models.
    """

    format: Literal["html", "markdown"] = Field(..., description="Format of the content.")
    content: str = Field(..., description="The actual page content.", examples=["<p><b>This</b> is rich text</p>"])


class CanvasPageContent(CodaBaseModel):
    """Canvas page content with type discriminator.

    Represents a page containing rich text/canvas content.
    Used in PageCreate for specifying initial page content.
    """

    type: Literal["canvas"] = Field(..., description="Indicates a page containing canvas content.")
    canvas_content: PageContent = Field(..., description="The canvas content.")


class EmbedPageContent(CodaBaseModel):
    """Embed page content with type discriminator.

    Represents a page that embeds external content.
    Used in PageCreate for embedding URLs.
    """

    type: Literal["embed"] = Field(..., description="Indicates a page that embeds other content.")
    url: str = Field(..., description="The URL of the content to embed.", examples=["https://example.com"])
    render_method: Literal["compatibility", "standard"] | None = Field(None, description="Render mode for the embed.")


# ============================================================================
# Page Creation & Update Models
# ============================================================================


class PageCreate(CodaBaseModel):
    """Payload for creating a new page in a doc.

    Can be used both for creating standalone pages and as initial_page
    when creating a new doc.
    """

    name: str | None = Field(None, description="Name of the page.", examples=["Launch Status"])
    subtitle: str | None = Field(
        None, description="Subtitle of the page.", examples=["See the status of launch-related tasks."]
    )
    icon_name: str | None = Field(None, description="Name of the icon.", examples=["rocket"])
    image_url: str | None = Field(
        None, description="Url of the cover image to use.", examples=["https://example.com/image.jpg"]
    )
    parent_page_id: str | None = Field(
        None,
        description="The ID of this new page's parent, if creating a subpage.",
        examples=["canvas-tuVwxYz"],
    )
    page_content: CanvasPageContent | EmbedPageContent | None = Field(
        None, description="Content to initialize the page with (canvas or embed)."
    )


class PageContentUpdate(CodaBaseModel):
    """Payload for updating the content of an existing page."""

    insertion_mode: Literal["append", "replace"] = Field(..., description="Mode for inserting content.")
    canvas_content: PageContent = Field(..., description="The canvas content to insert.")


class PageUpdate(CodaBaseModel):
    """Payload for updating a page."""

    name: str | None = Field(None, description="Name of the page.", examples=["Launch Status"])
    subtitle: str | None = Field(
        None, description="Subtitle of the page.", examples=["See the status of launch-related tasks."]
    )
    icon_name: str | None = Field(None, description="Name of the icon.", examples=["rocket"])
    image_url: str | None = Field(
        None, description="Url of the cover image to use.", examples=["https://example.com/image.jpg"]
    )
    is_hidden: bool | None = Field(
        None,
        description=(
            "Whether the page is hidden or not. Note that for pages that cannot be hidden, "
            "like the sole top-level page in a doc, this will be ignored."
        ),
        examples=[True],
    )
    content_update: PageContentUpdate | None = Field(None, description="Content with which to update an existing page.")


# ============================================================================
# Page Operation Results
# ============================================================================


class PageCreateResult(DocumentMutateResponse):
    """The result of a page creation."""

    id: str = Field(..., description="ID of the created page.", examples=["canvas-tuVwxYz"])


class PageUpdateResult(DocumentMutateResponse):
    """The result of a page update."""

    id: str = Field(..., description="ID of the updated page.", examples=["canvas-tuVwxYz"])


class PageDeleteResult(DocumentMutateResponse):
    """The result of a page deletion."""

    id: str = Field(..., description="ID of the page to be deleted.", examples=["canvas-tuVwxYz"])



"""Formula-related MCP tools for Coda API."""

from typing import Literal

from ..client import CodaClient, clean_params
from ..models import Formula, FormulaList, Method


async def list_formulas(
    client: CodaClient,
    doc_id: str,
    limit: int | None = None,
    page_token: str | None = None,
    sort_by: Literal["name"] | None = None,
) -> FormulaList:
    """List named formulas in a doc.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        sort_by: How to sort the results.

    Returns:
        List of named formulas with pagination metadata.
    """
    params = {"limit": limit, "pageToken": page_token, "sortBy": sort_by}
    result = await client.request(Method.GET, f"docs/{doc_id}/formulas", params=clean_params(params))
    return FormulaList.model_validate(result)


async def get_formula(client: CodaClient, doc_id: str, formula_id_or_name: str) -> Formula:
    """Get details about a specific formula.

    Args:
        client: The Coda client instance.
        doc_id: ID of the doc.
        formula_id_or_name: ID or name of the formula.

    Returns:
        Formula details including the computed value.
    """
    result = await client.request(Method.GET, f"docs/{doc_id}/formulas/{formula_id_or_name}")
    return Formula.model_validate(result)



"""Coda MCP server - main entry point and tool registration."""

from typing import Literal

from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP

from .client import CodaClient
from .models import (
    BeginPageContentExportRequest,
    BeginPageContentExportResponse,
    CanvasPageContent,
    Column,
    ColumnList,
    Doc,
    DocCreate,
    DocDelete,
    DocList,
    DocumentCreationResult,
    DocUpdate,
    DocUpdateResult,
    EmbedPageContent,
    Formula,
    FormulaList,
    Page,
    PageContentExportStatusResponse,
    PageContentUpdate,
    PageCreate,
    PageCreateResult,
    PageDeleteResult,
    PageList,
    PageUpdate,
    PageUpdateResult,
    PushButtonResult,
    Row,
    RowDeleteResult,
    RowEdit,
    RowList,
    RowsDeleteResult,
    RowsUpsertResult,
    RowUpdateResult,
    Table,
    TableList,
    User,
)
from .tools import docs, formulas, pages, rows, tables

# Central MCP server instance
load_dotenv()
mcp = FastMCP("coda", dependencies=["aiohttp"])
client = CodaClient()

# ============================================================================
# Doc Tools
# ============================================================================


@mcp.tool(
    description="Get information about the current authenticated Coda user including name, email, and scoped token info"
)
async def whoami() -> User:
    """Get information about the current authenticated user.

    Returns:
        User information including name, email, and scoped token info.
    """
    return await docs.whoami(client)


@mcp.tool(description="Get detailed metadata about a specific Coda doc by its ID")
async def get_doc_info(doc_id: str) -> Doc:
    """Get info about a particular doc."""
    return await docs.get_doc_info(client, doc_id)


@mcp.tool(description="Permanently delete a Coda doc by ID - use with extreme caution as this cannot be undone")
async def delete_doc(doc_id: str) -> DocDelete:
    """Delete a doc. USE WITH CAUTION."""
    return await docs.delete_doc(client, doc_id)


@mcp.tool(description="Update properties of a Coda doc including title and icon")
async def update_doc(doc_id: str, title: str | None = None, icon_name: str | None = None) -> DocUpdateResult:
    """Update properties of a doc."""
    request = DocUpdate(title=title, icon_name=icon_name)
    return await docs.update_doc(client, doc_id, request)


@mcp.tool(
    description=(
        "List Coda docs accessible by the user (defaults to your own unpublished docs) - "
        "returns docs in reverse chronological order by most recent activity"
    )
)
async def list_docs(
    is_owner: bool = True,
    is_published: bool = False,
    query: str | None = None,
    source_doc: str | None = None,
    is_starred: bool | None = None,
    in_gallery: bool | None = None,
    workspace_id: str | None = None,
    folder_id: str | None = None,
    limit: int | None = None,
    page_token: str | None = None,
) -> DocList:
    """List available docs.

    Returns a list of Coda docs accessible by the user, and which they have opened at least once.
    These are returned in the same order as on the docs page: reverse chronological by the latest
    event relevant to the user (last viewed, edited, or shared).

    Args:
        is_owner: Show only docs owned by the user (default: True).
        is_published: Show only published docs (default: False).
        query: Search term used to filter down results.
        source_doc: Show only docs copied from the specified doc ID.
        is_starred: If true, returns docs that are starred. If false, returns docs that are not starred.
        in_gallery: Show only docs visible within the gallery.
        workspace_id: Show only docs belonging to the given workspace.
        folder_id: Show only docs belonging to the given folder.
        limit: Maximum number of results to return in this query (default: 25).
        page_token: An opaque token used to fetch the next page of results.

    Returns:
        Dictionary containing document list and pagination info.
    """
    return await docs.list_docs(
        client,
        is_owner,
        is_published,
        query,
        source_doc,
        is_starred,
        in_gallery,
        workspace_id,
        folder_id,
        limit,
        page_token,
    )


@mcp.tool(
    description=(
        "Create a new Coda doc with optional configuration including title, timezone, "
        "folder placement, and initial page content"
    )
)
async def create_doc(
    title: str,
    source_doc: str | None = None,
    timezone: str | None = None,
    folder_id: str | None = None,
    initial_page: PageCreate | None = None,
) -> DocumentCreationResult:
    """Create a new Coda doc.

    Args:
        title: Title of the new doc.
        source_doc: Optional ID of a doc to copy.
        timezone: Timezone for the doc, e.g. 'America/Los_Angeles'.
        folder_id: ID of the folder to place the doc in.
        initial_page: Configuration for the initial page of the doc.
            Can include name, subtitle, iconName, imageUrl, parentPageId, and pageContent.

    Returns:
        DocumentCreationResult with the newly created doc's metadata.
    """
    request = DocCreate(
        title=title,
        source_doc=source_doc,
        timezone=timezone,
        folder_id=folder_id,
        initial_page=initial_page,
    )
    return await docs.create_doc(client, request)


# ============================================================================
# Page Tools
# ============================================================================


@mcp.tool(description="List all pages in a Coda doc with pagination support")
async def list_pages(
    doc_id: str,
    limit: int | None = None,
    page_token: str | None = None,
) -> PageList:
    """List pages in a Coda doc."""
    return await pages.list_pages(client, doc_id, limit, page_token)


@mcp.tool(description="Get detailed metadata about a specific page by its ID or name")
async def get_page(doc_id: str, page_id_or_name: str) -> Page:
    """Get details about a page."""
    return await pages.get_page(client, doc_id, page_id_or_name)


@mcp.tool(
    description=(
        "Update properties and content of a page including name, subtitle, icon, visibility, and HTML/markdown content"
    )
)
async def update_page(
    doc_id: str,
    page_id_or_name: str,
    name: str | None = None,
    subtitle: str | None = None,
    icon_name: str | None = None,
    image_url: str | None = None,
    is_hidden: bool | None = None,
    content_update: PageContentUpdate | None = None,
) -> PageUpdateResult:
    """Update properties of a page.

    Args:
        doc_id: The ID of the doc.
        page_id_or_name: The ID or name of the page.
        name: Name of the page.
        subtitle: Subtitle of the page.
        icon_name: Name of the icon.
        image_url: URL of the cover image.
        is_hidden: Whether the page is hidden.
        content_update: Content update payload, e.g.:
            {
                "insertion_mode": "append",
                "canvas_content": {
                    "format": "html",
                    "content": "<p><b>This</b> is rich text</p>"
                }
            }

    Returns:
        API response from Coda.
    """
    page_update = PageUpdate(
        name=name,
        subtitle=subtitle,
        icon_name=icon_name,
        image_url=image_url,
        is_hidden=is_hidden,
        content_update=content_update,
    )
    return await pages.update_page(client, doc_id, page_id_or_name, page_update)


@mcp.tool(description="Delete a page from a Coda doc by its ID or name")
async def delete_page(doc_id: str, page_id_or_name: str) -> PageDeleteResult:
    """Delete a page from a doc."""
    return await pages.delete_page(client, doc_id, page_id_or_name)


@mcp.tool(
    description=(
        "Start an async export of page content in HTML or markdown format - "
        "returns request ID to poll for completion with get_page_content_export_status"
    )
)
async def begin_page_content_export(
    doc_id: str, page_id_or_name: str, output_format: str = "html"
) -> BeginPageContentExportResponse:
    """Initiate an export of page content.

    This starts an asynchronous export process. The export is not immediate - you must poll
    the status using get_page_content_export_status with the returned request ID.

    IMPORTANT: Due to Coda's server replication, the export request may not be immediately
    available on all servers. If you get a 404 error when checking status, wait 2-3 seconds
    and retry with exponential backoff.

    Workflow:
    1. Call this endpoint to start export
    2. Wait 2-3 seconds for server replication
    3. Poll get_page_content_export_status until status="complete"
    4. Use the downloadLink from the status response to download content

    Args:
        doc_id: ID of the doc.
        page_id_or_name: ID or name of the page.
        output_format: Format for export - either "html" or "markdown".

    Returns:
        Export response with:
        - id: The request ID to use for polling status
        - status: Initial status (usually "inProgress")
        - href: URL to check export status
    """
    export_request = BeginPageContentExportRequest(output_format=output_format)
    return await pages.begin_page_content_export(client, doc_id, page_id_or_name, export_request)


@mcp.tool(
    description=(
        "Check status of a page export and auto-download content when ready - "
        "poll this after starting export with begin_page_content_export"
    )
)
async def get_page_content_export_status(
    doc_id: str, page_id_or_name: str, request_id: str
) -> PageContentExportStatusResponse:
    """Check the status of a page content export.

    Poll this endpoint to check if your export (initiated with begin_page_content_export) is ready.

    IMPORTANT: 404 errors are expected initially due to server replication lag. If you receive
    a 404 error, wait 2-3 seconds and retry. Use exponential backoff for subsequent retries.

    When the export completes, this function automatically downloads the content for you,
    so you receive the actual page content directly without needing to make an additional request.

    Args:
        doc_id: ID of the doc.
        page_id_or_name: ID or name of the page.
        request_id: The request ID returned from begin_page_content_export.

    Returns:
        Status response with:
        - id: The request ID
        - status: "inProgress", "complete", or "failed"
        - href: URL to check status again
        - downloadLink: (when status="complete") Temporary URL where content was downloaded from
        - content: (when status="complete") The actual exported page content (HTML or markdown)
        - error: (when status="failed") Error message describing what went wrong

    Next steps:
    - If status="inProgress": Wait 1-2 seconds and poll again
    - If status="complete": The content field contains the exported page content
    - If status="failed": Check error message and handle accordingly
    """
    return await pages.get_page_content_export_status(client, doc_id, page_id_or_name, request_id)


@mcp.tool(
    description=(
        "Create a new page in a Coda doc with optional subtitle, icon, parent page, and initial HTML/markdown content"
    )
)
async def create_page(
    doc_id: str,
    name: str,
    subtitle: str | None = None,
    icon_name: str | None = None,
    image_url: str | None = None,
    parent_page_id: str | None = None,
    page_content: CanvasPageContent | EmbedPageContent | None = None,
) -> PageCreateResult:
    """Create a new page in a doc.

    Args:
        doc_id: The ID of the doc.
        name: Name of the page.
        subtitle: Subtitle of the page.
        icon_name: Name of the icon.
        image_url: URL of the cover image.
        parent_page_id: The ID of this new page's parent, if creating a subpage.
        page_content: Content to initialize the page with (canvas or embed), e.g.:
            {
                "type": "canvas",
                "canvas_content": {
                    "format": "html",
                    "content": "<p><b>This</b> is rich text</p>"
                }
            }

    Returns:
        API response from Coda.
    """
    page_create = PageCreate(
        name=name,
        subtitle=subtitle,
        icon_name=icon_name,
        image_url=image_url,
        parent_page_id=parent_page_id,
        page_content=page_content,
    )
    return await pages.create_page(client, doc_id, page_create)


# ============================================================================
# Table Tools
# ============================================================================


@mcp.tool(description="List all tables and views in a Coda doc with optional filtering and sorting")
async def list_tables(
    doc_id: str,
    limit: int | None = None,
    page_token: str | None = None,
    sort_by: Literal["name"] | None = None,
    table_types: list[str] | None = None,
) -> TableList:
    """List tables in a Coda doc.

    Args:
        doc_id: ID of the doc.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        sort_by: How to sort the results (e.g., 'name').
        table_types: Types of tables to include (e.g., ['table', 'view']).

    Returns:
        List of tables with their metadata.
    """
    return await tables.list_tables(client, doc_id, limit, page_token, sort_by, table_types)


@mcp.tool(description="Get detailed information about a specific table including its schema, columns, and metadata")
async def get_table(doc_id: str, table_id_or_name: str) -> Table:
    """Get details about a specific table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.

    Returns:
        Table details including columns and metadata.
    """
    return await tables.get_table(client, doc_id, table_id_or_name)


@mcp.tool(description="List all columns in a table with their properties, formats, and formulas")
async def list_columns(
    doc_id: str,
    table_id_or_name: str,
    limit: int | None = None,
    page_token: str | None = None,
    visible_only: bool | None = None,
) -> ColumnList:
    """List columns in a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        visible_only: If true, only return visible columns.

    Returns:
        List of columns with their properties.
    """
    return await tables.list_columns(client, doc_id, table_id_or_name, limit, page_token, visible_only)


@mcp.tool(description="Get detailed information about a specific column including its type, format, and formula")
async def get_column(doc_id: str, table_id_or_name: str, column_id_or_name: str) -> Column:
    """Get details about a specific column.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        column_id_or_name: ID or name of the column.

    Returns:
        Column details including format and formula.
    """
    return await tables.get_column(client, doc_id, table_id_or_name, column_id_or_name)


@mcp.tool(
    description=(
        "Trigger a button column in a table row to execute its automation or action "
        "(buttons can run formulas, modify data, or trigger workflows)"
    )
)
async def push_button(
    doc_id: str,
    table_id_or_name: str,
    row_id_or_name: str,
    column_id_or_name: str,
) -> PushButtonResult:
    """Push a button in a table cell.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID of the table.
        row_id_or_name: ID of the row containing the button.
        column_id_or_name: ID of the column containing the button.

    Returns:
        Result of the button push operation.
    """
    return await tables.push_button(client, doc_id, table_id_or_name, row_id_or_name, column_id_or_name)


# ============================================================================
# Row Tools
# ============================================================================


@mcp.tool(
    description=(
        "List rows in a table with optional filtering, sorting, and pagination - returns row data with cell values"
    )
)
async def list_rows(
    doc_id: str,
    table_id_or_name: str,
    query: str | None = None,
    sort_by: str | None = None,
    use_column_names: bool | None = None,
    value_format: Literal["simple", "simpleWithArrays", "rich"] | None = None,
    visible_only: bool | None = None,
    limit: int | None = None,
    page_token: str | None = None,
    sync_token: str | None = None,
) -> RowList:
    """List rows in a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        query: Query to filter rows (e.g., 'Status="Complete"').
        sort_by: Column to sort by. Use 'natural' for the table's sort order.
        use_column_names: Use column names instead of IDs in the response.
        value_format: Format for cell values (simple, simpleWithArrays, or rich).
        visible_only: If true, only return visible rows.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        sync_token: Token for incremental sync of changes.

    Returns:
        List of rows with their values.
    """
    return await rows.list_rows(
        client,
        doc_id,
        table_id_or_name,
        query,
        sort_by,
        use_column_names,
        value_format,
        visible_only,
        limit,
        page_token,
        sync_token,
    )


@mcp.tool(description="Get a specific row from a table by its ID or name with all cell values")
async def get_row(
    doc_id: str,
    table_id_or_name: str,
    row_id_or_name: str,
    use_column_names: bool | None = None,
    value_format: Literal["simple", "simpleWithArrays", "rich"] | None = None,
) -> Row:
    """Get a specific row from a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row.
        use_column_names: Use column names instead of IDs in the response.
        value_format: Format for cell values (simple, simpleWithArrays, or rich).

    Returns:
        Row data with values.
    """
    return await rows.get_row(client, doc_id, table_id_or_name, row_id_or_name, use_column_names, value_format)


@mcp.tool(
    description="Insert new rows or update existing rows in a table based on key columns - ideal for bulk operations"
)
async def upsert_rows(
    doc_id: str,
    table_id_or_name: str,
    rows_data: list[RowEdit],
    key_columns: list[str] | None = None,
    disable_parsing: bool | None = None,
) -> RowsUpsertResult:
    """Insert or update rows in a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        rows_data: List of rows to upsert. Each row should have a 'cells' array with column/value pairs.
        key_columns: Column IDs/names to use as keys for matching existing rows.
        disable_parsing: If true, cell values won't be parsed (e.g., URLs won't become links).

    Returns:
        Result of the upsert operation.
    """
    return await rows.upsert_rows(client, doc_id, table_id_or_name, rows_data, key_columns, disable_parsing)


@mcp.tool(description="Update cell values in a specific row by its ID or name")
async def update_row(
    doc_id: str,
    table_id_or_name: str,
    row_id_or_name: str,
    row: RowEdit,
    disable_parsing: bool | None = None,
) -> RowUpdateResult:
    """Update a specific row in a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row to update.
        row: Row data with cells array containing column/value pairs.
        disable_parsing: If true, cell values won't be parsed.

    Returns:
        Updated row data.
    """
    return await rows.update_row(client, doc_id, table_id_or_name, row_id_or_name, row, disable_parsing)


@mcp.tool(description="Delete a specific row from a table by its ID or name")
async def delete_row(doc_id: str, table_id_or_name: str, row_id_or_name: str) -> RowDeleteResult:
    """Delete a specific row from a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_id_or_name: ID or name of the row to delete.

    Returns:
        Result of the deletion.
    """
    return await rows.delete_row(client, doc_id, table_id_or_name, row_id_or_name)


@mcp.tool(description="Delete multiple rows from a table at once using a list of row IDs")
async def delete_rows(
    doc_id: str,
    table_id_or_name: str,
    row_ids: list[str],
) -> RowsDeleteResult:
    """Delete multiple rows from a table.

    Args:
        doc_id: ID of the doc.
        table_id_or_name: ID or name of the table.
        row_ids: List of row IDs to delete.

    Returns:
        Result of the deletion operation.
    """
    return await rows.delete_rows(client, doc_id, table_id_or_name, row_ids)


# ============================================================================
# Formula Tools
# ============================================================================


@mcp.tool(description="List all named formulas in a Coda doc with their names and IDs")
async def list_formulas(
    doc_id: str,
    limit: int | None = None,
    page_token: str | None = None,
    sort_by: Literal["name"] | None = None,
) -> FormulaList:
    """List named formulas in a doc.

    Args:
        doc_id: ID of the doc.
        limit: Maximum number of results to return.
        page_token: An opaque token to fetch the next page of results.
        sort_by: How to sort the results.

    Returns:
        List of named formulas with pagination metadata.
    """
    return await formulas.list_formulas(client, doc_id, limit, page_token, sort_by)


@mcp.tool(description="Get details about a specific named formula including its computed value")
async def get_formula(doc_id: str, formula_id_or_name: str) -> Formula:
    """Get details about a specific formula.

    Args:
        doc_id: ID of the doc.
        formula_id_or_name: ID or name of the formula.

    Returns:
        Formula details including the computed value.
    """
    return await formulas.get_formula(client, doc_id, formula_id_or_name)


# ============================================================================
# Server Entry Point
# ============================================================================


def main() -> None:
    """Run the server."""
    mcp.run()


if __name__ == "__main__":
    main()



"""Pydantic models for Coda page content export API.

This module contains models for the asynchronous page content export workflow:
1. Begin export with BeginPageContentExportRequest -> BeginPageContentExportResponse
2. Poll status with PageContentExportStatusResponse
3. Download content from downloadLink when status is "complete"

The export API is asynchronous because rendering page content can take time.
Poll the href URL from the initial response to check status and get the download link.
"""

from typing import Literal

from pydantic import Field

from .common import CodaBaseModel


class BeginPageContentExportRequest(CodaBaseModel):
    """Request for beginning an export of page content.

    The export process is asynchronous - this request initiates the export
    and returns a status URL to poll for completion.
    """

    output_format: Literal["html", "markdown"] = Field(
        ...,
        description="Supported output content formats that can be requested for getting content for an existing page.",
    )


class BeginPageContentExportResponse(CodaBaseModel):
    """Response when beginning an export of page content.

    Contains the export request ID and a status URL to poll for completion.
    The export happens asynchronously - poll the href URL to check status
    and get the download link when complete.
    """

    id: str = Field(
        ...,
        description="The identifier of this export request.",
        examples=["AbCDeFGH"],
    )
    status: str = Field(
        ...,
        description="The status of this export.",
        examples=["complete"],
    )
    href: str = Field(
        ...,
        description=(
            "The URL that reports the status of this export. Poll this URL to get "
            "the content URL when the export has completed."
        ),
        examples=["https://coda.io/apis/v1/docs/somedoc/pages/somepage/export/some-request-id"],
    )


class PageContentExportStatusResponse(CodaBaseModel):
    """Response when requesting the status of a page content export.

    Poll this endpoint to check export status. When status is "complete",
    the downloadLink field will contain the URL to download the exported content.
    The download link typically expires after a short time - call this endpoint
    again to get a fresh link if needed.

    Status values:
    - "inProgress": Export is still being processed
    - "complete": Export finished successfully, downloadLink is available
    - "failed": Export failed, error field contains details
    """

    id: str = Field(
        ...,
        description="The identifier of this export request.",
        examples=["AbCDeFGH"],
    )
    status: Literal["inProgress", "failed", "complete"] = Field(
        ...,
        description="The status of this export.",
        examples=["complete"],
    )
    href: str = Field(
        ...,
        description="The URL that reports the status of this export.",
        examples=["https://coda.io/apis/v1/docs/somedoc/pages/somepage/export/some-request-id"],
    )
    download_link: str | None = Field(
        None,
        description=(
            "Once the export completes, the location where the resulting export "
            "file can be downloaded; this link typically expires after a short "
            "time. Call this method again to get a fresh link."
        ),
        examples=["https://coda.io/blobs/DOC_EXPORT_RENDERING/some-request-id"],
    )
    error: str | None = Field(
        None,
        description="Message describing an error, if this export failed.",
    )
    content: str | None = Field(
        None,
        description=(
            "The actual exported page content (HTML or markdown). "
            "This is automatically downloaded when status is 'complete'. "
            "Note: This field is a convenience enhancement not in the OpenAPI spec."
        ),
    )



"""Tests for CodaClient initialization."""

from pytest import MonkeyPatch

from coda_mcp_server.client import CodaClient


class TestCodaClient:
    """Test the CodaClient class initialization."""

    def test_init_with_explicit_token(self, monkeypatch: MonkeyPatch) -> None:
        """Test client initialization with an explicit API token."""
        # Clear environment variable to ensure we're testing explicit token
        monkeypatch.delenv("CODA_API_KEY", raising=False)

        client = CodaClient(api_token="explicit-token-123")

        assert client.api_token == "explicit-token-123"
        assert client.base_url == "https://coda.io/apis/v1"
        assert client.headers["Authorization"] == "Bearer explicit-token-123"
        assert client.headers["Content-Type"] == "application/json"

    def test_init_with_env_variable(self, monkeypatch: MonkeyPatch) -> None:
        """Test client initialization with environment variable."""
        # Set environment variable
        monkeypatch.setenv("CODA_API_KEY", "env-token-456")

        client = CodaClient()

        assert client.api_token == "env-token-456"
        assert client.headers["Authorization"] == "Bearer env-token-456"

    def test_init_env_variable_takes_precedence(self, monkeypatch: MonkeyPatch) -> None:
        """Test that environment variable takes precedence over explicit token."""
        # Set environment variable
        monkeypatch.setenv("CODA_API_KEY", "env-token-789")

        # Pass explicit token
        client = CodaClient(api_token="explicit-token-000")

        # Environment variable should be used
        assert client.api_token == "env-token-789"
        assert client.headers["Authorization"] == "Bearer env-token-789"

    def test_init_without_token(self, monkeypatch: MonkeyPatch) -> None:
        """Test client initialization without any token."""
        # Clear environment variable
        monkeypatch.delenv("CODA_API_KEY", raising=False)

        client = CodaClient()

        assert client.api_token is None
        assert client.headers["Authorization"] == "Bearer None"  # This would fail in real API calls

    def test_headers_structure(self) -> None:
        """Test that headers are properly structured."""
        client = CodaClient(api_token="test-token")

        assert isinstance(client.headers, dict)
        assert "Authorization" in client.headers
        assert "Content-Type" in client.headers
        assert len(client.headers) == 2

    def test_base_url_format(self) -> None:
        """Test that base URL is correctly formatted."""
        client = CodaClient(api_token="test-token")

        assert client.base_url.startswith("https://")
        assert "coda.io" in client.base_url
        assert client.base_url.endswith("/apis/v1")
        assert not client.base_url.endswith("/")  # No trailing slash



"""Tests for Pydantic models and snake_case serialization."""

from coda_mcp_server.models import (
    Doc,
    DocList,
    Page,
    PageList,
    Row,
    Table,
    User,
)
from coda_mcp_server.models.common import normalize_keys


class TestNormalizeKeys:
    """Test the normalize_keys utility function."""

    def test_to_snake_simple(self) -> None:
        """Test conversion of simple camelCase to snake_case."""
        data = {"firstName": "John", "lastName": "Doe"}
        result = normalize_keys(data, "to_snake")
        assert result == {"first_name": "John", "last_name": "Doe"}

    def test_to_snake_nested(self) -> None:
        """Test conversion of nested objects."""
        data = {
            "userData": {"firstName": "John", "lastName": "Doe"},
            "accountInfo": {"accountId": "123", "accountType": "premium"},
        }
        result = normalize_keys(data, "to_snake")
        assert result == {
            "user_data": {"first_name": "John", "last_name": "Doe"},
            "account_info": {"account_id": "123", "account_type": "premium"},
        }

    def test_to_snake_list(self) -> None:
        """Test conversion of lists of objects."""
        data = {"items": [{"itemName": "Item1"}, {"itemName": "Item2"}]}
        result = normalize_keys(data, "to_snake")
        assert result == {"items": [{"item_name": "Item1"}, {"item_name": "Item2"}]}

    def test_to_camel_simple(self) -> None:
        """Test conversion of simple snake_case to camelCase."""
        data = {"first_name": "John", "last_name": "Doe"}
        result = normalize_keys(data, "to_camel")
        assert result == {"firstName": "John", "lastName": "Doe"}

    def test_to_camel_nested(self) -> None:
        """Test conversion of nested objects to camelCase."""
        data = {
            "user_data": {"first_name": "John", "last_name": "Doe"},
            "account_info": {"account_id": "123", "account_type": "premium"},
        }
        result = normalize_keys(data, "to_camel")
        assert result == {
            "userData": {"firstName": "John", "lastName": "Doe"},
            "accountInfo": {"accountId": "123", "accountType": "premium"},
        }

    def test_preserves_non_string_keys(self) -> None:
        """Test that non-string keys are preserved."""
        data = {1: "one", 2: "two", "normalKey": "value"}
        result = normalize_keys(data, "to_snake")
        assert result == {1: "one", 2: "two", "normal_key": "value"}


class TestCodaBaseModel:
    """Test the CodaBaseModel base class."""

    def test_accepts_camel_case_input(self) -> None:
        """Test that camelCase input is automatically normalized to snake_case."""
        data = {
            "name": "Test User",
            "loginId": "test@example.com",  # camelCase
            "type": "user",
            "scoped": True,
            "tokenName": "test-token",  # camelCase
            "href": "https://coda.io/apis/v1/whoami",
            "workspace": {
                "id": "ws-123",
                "type": "workspace",
                "organizationId": "org-456",  # camelCase
                "browserLink": "https://coda.io/docs",  # camelCase
                "name": "Test Workspace",
            },
        }
        user = User.model_validate(data)
        assert user.login_id == "test@example.com"
        assert user.token_name == "test-token"
        assert user.workspace.organization_id == "org-456"
        assert user.workspace.browser_link == "https://coda.io/docs"

    def test_accepts_snake_case_input(self) -> None:
        """Test that snake_case input also works."""
        data = {
            "name": "Test User",
            "login_id": "test@example.com",  # snake_case
            "type": "user",
            "scoped": True,
            "token_name": "test-token",  # snake_case
            "href": "https://coda.io/apis/v1/whoami",
            "workspace": {
                "id": "ws-123",
                "type": "workspace",
                "organization_id": "org-456",  # snake_case
                "browser_link": "https://coda.io/docs",  # snake_case
                "name": "Test Workspace",
            },
        }
        user = User.model_validate(data)
        assert user.login_id == "test@example.com"
        assert user.token_name == "test-token"

    def test_model_dump_returns_snake_case(self) -> None:
        """Test that model_dump() returns snake_case by default."""
        data = {
            "name": "Test User",
            "loginId": "test@example.com",
            "type": "user",
            "scoped": True,
            "tokenName": "test-token",
            "href": "https://coda.io/apis/v1/whoami",
            "workspace": {
                "id": "ws-123",
                "type": "workspace",
                "browserLink": "https://coda.io/docs",
                "name": "Test Workspace",
            },
        }
        user = User.model_validate(data)
        dumped = user.model_dump()

        # Verify top-level fields are snake_case
        assert "login_id" in dumped
        assert "token_name" in dumped
        assert "loginId" not in dumped
        assert "tokenName" not in dumped

        # Verify nested fields are snake_case
        assert "browser_link" in dumped["workspace"]
        assert "browserLink" not in dumped["workspace"]

    def test_model_dump_camel_returns_camel_case(self) -> None:
        """Test that model_dump_camel() returns camelCase for API."""
        data = {
            "name": "Test User",
            "login_id": "test@example.com",
            "type": "user",
            "scoped": True,
            "token_name": "test-token",
            "href": "https://coda.io/apis/v1/whoami",
            "workspace": {
                "id": "ws-123",
                "type": "workspace",
                "organization_id": "org-456",
                "browser_link": "https://coda.io/docs",
                "name": "Test Workspace",
            },
        }
        user = User.model_validate(data)
        dumped = user.model_dump_camel()

        # Verify top-level fields are camelCase
        assert "loginId" in dumped
        assert "tokenName" in dumped
        assert "login_id" not in dumped
        assert "token_name" not in dumped

        # Verify nested fields are camelCase
        assert "browserLink" in dumped["workspace"]
        assert "organizationId" in dumped["workspace"]
        assert "browser_link" not in dumped["workspace"]


class TestDocModels:
    """Test Doc-related models."""

    def test_doc_model_validates(self) -> None:
        """Test that Doc model can be created and validated."""
        data = {
            "id": "test-doc-123",
            "type": "doc",
            "href": "https://coda.io/apis/v1/docs/test-doc-123",
            "browserLink": "https://coda.io/d/_dtest-doc-123",
            "name": "Test Doc",
            "owner": "test@example.com",
            "ownerName": "Test User",
            "createdAt": "2025-01-01T00:00:00.000Z",
            "updatedAt": "2025-01-01T00:00:00.000Z",
            "workspace": {
                "id": "ws-123",
                "type": "workspace",
                "browserLink": "https://coda.io/docs",
                "name": "Test Workspace",
            },
            "folder": {
                "id": "fl-123",
                "type": "folder",
                "browserLink": "https://coda.io/docs?folderId=fl-123",
                "name": "Test Folder",
            },
            "workspaceId": "ws-123",
            "folderId": "fl-123",
        }
        doc = Doc.model_validate(data)
        assert doc.id == "test-doc-123"
        assert doc.owner_name == "Test User"
        assert doc.browser_link == "https://coda.io/d/_dtest-doc-123"
        assert doc.workspace_id == "ws-123"
        assert doc.folder_id == "fl-123"

    def test_doc_list_validates(self) -> None:
        """Test that DocList model can be created."""
        data = {
            "items": [
                {
                    "id": "doc1",
                    "type": "doc",
                    "href": "https://coda.io/apis/v1/docs/doc1",
                    "browserLink": "https://coda.io/d/_ddoc1",
                    "name": "Doc 1",
                    "owner": "test@example.com",
                    "ownerName": "Test User",
                    "createdAt": "2025-01-01T00:00:00.000Z",
                    "updatedAt": "2025-01-01T00:00:00.000Z",
                    "workspace": {
                        "id": "ws-123",
                        "type": "workspace",
                        "browserLink": "https://coda.io/docs",
                        "name": "Test Workspace",
                    },
                    "folder": {
                        "id": "fl-123",
                        "type": "folder",
                        "browserLink": "https://coda.io/docs?folderId=fl-123",
                        "name": "Test Folder",
                    },
                    "workspaceId": "ws-123",
                    "folderId": "fl-123",
                }
            ],
            "href": "https://coda.io/apis/v1/docs",
        }
        doc_list = DocList.model_validate(data)
        assert len(doc_list.items) == 1
        assert doc_list.items[0].owner_name == "Test User"


class TestPageModels:
    """Test Page-related models."""

    def test_page_model_validates(self) -> None:
        """Test that Page model can be created."""
        data = {
            "id": "canvas-test123",
            "type": "page",
            "href": "https://coda.io/apis/v1/docs/doc123/pages/canvas-test123",
            "browserLink": "https://coda.io/d/_ddoc123/_sutest123",
            "name": "Test Page",
            "subtitle": "",
            "contentType": "canvas",
            "isHidden": False,
            "isEffectivelyHidden": False,
            "children": [],
        }
        page = Page.model_validate(data)
        assert page.id == "canvas-test123"
        assert page.content_type == "canvas"
        assert page.is_hidden is False
        assert page.browser_link == "https://coda.io/d/_ddoc123/_sutest123"

    def test_page_list_validates(self) -> None:
        """Test that PageList model can be created."""
        data = {
            "items": [
                {
                    "id": "canvas-test123",
                    "type": "page",
                    "href": "https://coda.io/apis/v1/docs/doc123/pages/canvas-test123",
                    "browserLink": "https://coda.io/d/_ddoc123/_sutest123",
                    "name": "Test Page",
                    "contentType": "canvas",
                    "isHidden": False,
                    "isEffectivelyHidden": False,
                    "children": [],
                }
            ],
            "href": "https://coda.io/apis/v1/docs/doc123/pages",
        }
        page_list = PageList.model_validate(data)
        assert len(page_list.items) == 1


class TestTableModels:
    """Test Table-related models."""

    def test_table_model_validates(self) -> None:
        """Test that Table model can be created."""
        data = {
            "id": "grid-test123",
            "type": "table",
            "tableType": "table",
            "href": "https://coda.io/apis/v1/docs/doc123/tables/grid-test123",
            "browserLink": "https://coda.io/d/_ddoc123#_tugrid-test123",
            "name": "Test Table",
            "parent": {
                "id": "canvas-page123",
                "type": "page",
                "href": "https://coda.io/apis/v1/docs/doc123/pages/canvas-page123",
                "browserLink": "https://coda.io/d/_ddoc123/_supage123",
                "name": "Test Page",
            },
            "rowCount": 10,
            "createdAt": "2025-01-01T00:00:00.000Z",
            "updatedAt": "2025-01-01T00:00:00.000Z",
            "displayColumn": {
                "id": "c-abc123",
                "type": "column",
                "href": "https://coda.io/apis/v1/docs/doc123/tables/grid-test123/columns/c-abc123",
            },
            "sorts": [],
            "layout": "default",
        }
        table = Table.model_validate(data)
        assert table.id == "grid-test123"
        assert table.table_type == "table"
        assert table.row_count == 10


class TestRowModels:
    """Test Row-related models."""

    def test_row_model_validates(self) -> None:
        """Test that Row model can be created."""
        data = {
            "id": "i-test123",
            "type": "row",
            "href": "https://coda.io/apis/v1/docs/doc123/tables/grid-abc/rows/i-test123",
            "name": "Test Row",
            "index": 0,
            "browserLink": "https://coda.io/d/_ddoc123#_tugrid-abc/_rui-test123",
            "createdAt": "2025-01-01T00:00:00.000Z",
            "updatedAt": "2025-01-01T00:00:00.000Z",
            "values": {"col1": "value1", "col2": "value2"},
        }
        row = Row.model_validate(data)
        assert row.id == "i-test123"
        assert row.browser_link == "https://coda.io/d/_ddoc123#_tugrid-abc/_rui-test123"
        assert row.created_at is not None
        assert row.updated_at is not None
